;
;  EXAMPLE.ASM	Also see SAMPLE.FOR for the F77L calls to these examples.
;
;  example:	This routine demonstrates how a function written
;		entirely in assembly language accesses arguments from,
;		and returns values to, a calling program written in
;		Lahey Fortran
;
;  make_lower:	A utility that makes a character string lower case and
;		demonstrates accessing the character type.
;
;  io_out:	Output a byte to a port


INT2SIZE equ 2

; It is very important when linking to F77L program units that any data
; segments your assembly code uses have their class name as 'DATA' in order
; to link correctly.  In addition, you may also choose to use the GROUP
; directive for addressing convenience if you will have multiple data
; segments.  Using the GROUP directive, and naming the data group DGROUP,
; also makes your data part of Fortran's data group and allows use of a number
; of F77L subroutines which either assume or set ds equal to dgroup.  In
; either case, take care that the value you actually load into the ds register
; during execution matches the value specified to the assembler in the ASSUME
; directive.
;
; It is up to the assembly code to load ds accordingly whenever it attempts
; to reference its data area (see code in 'example', below).
;
; In a group directive, never include a name of a segment which doesn't exist.
; Although the assembler and linker will not catch this error, your program
; will behave incorrectly.
;
dgroup	group	data
data	segment	word public 'DATA'
public	called	  ;public allows 'called' to be accessible to other modules

called	dw	0
data	ends
;
; Similarly, any code segments should either have a class name of 'CODE',
; or not specified (null).
;
excode	segment	byte
assume	cs:excode, ds:dgroup
;
; The procedure names must be declared public to be addressable by
; other modules.
;
public	example, make_lower, io_out

; log4 = example( int2, int4, int2a, int2a(3), dp, cmpx )
;		  arg1  arg2  arg3     arg4   arg5 arg6
;
example	proc	far
;
; Upon entry to any subroutine, F77L has pushed addresses of
; each argument onto the stack.  In the calling program,
; 'example' was defined as a function call rather than a simple
; subroutine; in this case, the compiler will also push the function's
; return value address onto the stack.  When all arguments have
; been dealt with, F77L issues a far call to the named routine.
;
	push	bp		;always do this...
	mov	bp, sp		;and this (bp must be preserved)
;
; The arguments are pushed right to left; thus, the first argument
; is closest to bp.  At this point, the stack contains:
;
;	segment of argn
;	offset of argn
;	     ...
;	segment of arg1
;	offset of arg1
;	segment of function return value
;	offset of function return value
;	return code segment (CS) }
;	return code offset  (IP) }-- four byte address from FAR call
;	saved bp	<-- BP, SP
;
; The first argument address is now 10 bytes from BP, referenced at [BP+10]
; and [BP+12].  Note that the minimum argument displacement is 6 for
; subroutines and 10 for functions.
;
; If this routine needed any temporary storage for "local" variables, it
; could allocate it now from the stack by adjusting the stack pointer sp
; down by the number of bytes needed, and referencing the space off of bp
; in a negative direction.  See Appendix E for more information if your
; routine needs more than 240 bytes.
;
	sub	sp, 30			;allocate 30 bytes for 'example'
	mov	word ptr [bp-2], 3	;loads a 3 into the top word of
					;local storage

	lds	bx, dword ptr [bp+10]	;point to arg1 (int*2)
	mov	ax, [bx]		;load it
	cwd				;convert to int*4 for addition
	lds	bx, dword ptr [bp+14]	;point to arg2 (int*4)
	add	[bx], ax
	adc	[bx+2], dx		;store sum in arg2
;
; Arg1 is added to arg2, and the result is returned in arg2.
;
	fild	dword ptr [bx]		;load int*4 onto NDP chip
	lds	bx, [bp+26]		;point to double precision arg5
	fstp	qword ptr [bx]		;store the 8-byte dp value
;
; A data type conversion made easy using the math coprocessor
;
; Arguments 3 & 4 in this example refer to array elements.  Since only
; addresses are passed, note that arguments of 'i2array' and 'i2array(1)'
; will pass the same address values.  If the array had been declared as
; i2array(5:8), then 'i2array' and 'i2array(5)' would pass the same values,
; and so on.  Specifying 'i2array' is a convenient way to conceptually
; pass an entire array, while 'i2array(n)' might be used to pass one element,
; or a subset, of an array.
;
	lds	bx, dword ptr [bp+18]	;point to arg3
	mov	ax, [bx+0*INT2SIZE]	;this fetches the first array element
	mov	ax, [bx+2*INT2SIZE]	;this fetches the third, etc.
	lds	bx, dword ptr [bp+22]	;in this example, the third array
	mov	cx, [bx]		;element is also available in arg4
;
; The registers ax and cx will contain the same value
;
	lds	bx, dword ptr [bp+30]	;point to complex arg6
	fld	dword ptr [bx]		;this loads the real part
	fld	st(0)			;duplicate on stack
	fmul	st(1), st		;square real part
	fld	dword ptr [bx+4]	;this loads the imaginary part
	fmul	st(1), st
	fmul	st(0), st		;square imaginary part
	fsubp	st(2), st
	fadd	st(0), st	;result: st(1) new real, st(0) imaginary
	fstp	dword ptr [bx+4]	;these store the squared value in arg6
	fstp	dword ptr [bx]		;and empty the NDP stack
;
; Complex variables are stored in memory with the real part first, at the
; pointer address, then the imaginary part 4 bytes subsequent for COMPLEX,
; and 8 bytes subsequent for COMPLEX*16.  The above code squares the input
; argument and returns the value in that argument.
;
; If your assembler code uses the NDP chip for any operations, you should make
; sure that its stack is left empty before returning to F77L.
;
	mov	ax, dgroup
	mov	ds, ax		;set ds accordingly to reference data group
	inc	called		;this value could also be used by another
				;external routine (was declared public)
	xor	al, al
	cmp	called, 1
	jne	ex1
	inc	al
ex1:	lds	bx, dword ptr [bp+6]	;point to function's return value
	mov	[bx], al	;this stores return into caller's variable
;
; Example was declared as a LOGICAL*4 function.  The above code returns
; .TRUE. if this is the first time example was called, else .FALSE.; F77L
; uses a zero value for false, and one for true.  Throughout the compiler,
; only the first (lowest address) byte of a logical variable is checked --
; the extra three bytes of a LOGICAL*4 are irrelevant.
;
; Routine is done -- now exit.  The assembler routine MUST restore the
; original bp and sp (as shown below), and clear the direction flag (cld)
; if it was set anywhere within.
;
	mov	sp, bp		;restore stack frame (this is required to
				;safely deallocate any local storage used)
	pop	bp		;and caller's bp
	ret			;FAR return
;
; The calling routine is responsible for removing the argument pointers
; from the stack, not the subroutine!
;
example	endp


; make_lower turns a character string into all lower case.  The argument
; it receives is a pointer to a character variable descriptor structure,
; NOT the variable itself.  The descriptor contains the variable's address
; and its length.
;
; Upon entry to any routine, the direction flag is clear; thus, the lodsb
; instruction below will auto-increment the pointer register.
;
make_lower	proc far
	push	bp
	mov	bp, sp
	lds	bx, dword ptr [bp+6]	;point to the character structure
	mov	cx, [bx+4]		;get the length from the structure
	lds	si, dword ptr [bx]	;now point to the first character value
next:	lodsb				;get next character
	cmp	al, 'A'
	jb	cycle
	cmp	al, 'Z'
	ja	cycle
	or	al, 20h			;it's an upper case letter, so lower it
	mov	[si-1], al		;and store back into the string
cycle:	loop	next			;do this 'length' times
	pop	bp
	ret
make_lower	endp


io_out	proc far
	push	bp
	mov	bp, sp
	lds	bx, dword ptr [bp+6]
	mov	dx, word ptr [bx]	;load port value (1st argument)
	lds	bx, dword ptr [bp+10]
	mov	al, byte ptr [bx]	;load byte value (2nd argument)
	out	dx, al			;output a byte
	pop	bp
	ret
io_out	endp

excode	ends

	end
;
; Do not put any label name or expression after the end statement!  This
; defines a program entry point, which has already been defined by the
; Fortran MAIN module.

